
#include <SPI.h>

#define SERIAL_BAUD          115200
#define SPI_BAUD             500000
#define DEBUG_PULSE_MICROS   50

// Attention: Pin10 (PB2/SS) is a bit specitial and must not become input after SPI was initialized
#define PIN_CS_RFM    10
#define PIN_CS_FLASH  8
#define PIN_DEBUG     A2


#define MODE_SLEEP          0x00
#define MODE_STANDBY        0x04
#define MODE_TX             0x0C


void debugPulse(int count)
{
  for (int i=0; i<count; i++)
  {
    if (i>0)
    {
      delayMicroseconds(DEBUG_PULSE_MICROS);
    }
    digitalWrite(PIN_DEBUG, HIGH);
    delayMicroseconds(DEBUG_PULSE_MICROS);
    digitalWrite(PIN_DEBUG, LOW);
  }
}


byte spiReadWrite(byte reg, byte val, uint8_t cs_pin)
{
  byte ret;
  digitalWrite(cs_pin, LOW);
  SPI.transfer(reg);
  ret = SPI.transfer(val);
  digitalWrite(cs_pin, HIGH);
  return ret;
}

void rfmWriteReg(byte reg, byte val)
{
  spiReadWrite(reg | 0x80, val, PIN_CS_RFM);
}

byte rfmReadReg(byte reg)
{
  return spiReadWrite(reg & 0x7F, 0x00, PIN_CS_RFM);
}


// set mode, and wait for completion (polling ModeReady, should be fairly quick)
void setOperationMode(byte mode)
{
  rfmWriteReg(0x01, mode); //set OP-MODE
  while ((rfmReadReg(0x27) & 0x80) == 0)
  {
  }
}

void initialize_OOK()
{
  rfmWriteReg(0x01, 0x04); // set OP-MODE (STANDBY)
  rfmWriteReg(0x11, 0x84); // set PaLevel (reduce OutputPower)
  rfmWriteReg(0x02, 0x08); // set DATA-Mode = Packet/OOK
  rfmWriteReg(0x2D, 0x00); // set PreambleLsb = 0
  rfmWriteReg(0x2E, 0x00); // set SyncConfig (SYNC = off)
  rfmWriteReg(0x37, 0x80); // set PacketConfig1 (VARIABLE, noCRC, no filter)
  rfmWriteReg(0x38, 0x00); // set PayloadLength (ignored in RX, max in TX)
  rfmWriteReg(0x25, 0x44); // set DioMapping1 (in Packet-TX: Dio0: TxReady, Dio2: Data)
  rfmWriteReg(0x07, 0x6C); // set (carrier) Frequency MSB
  rfmWriteReg(0x08, 0x77); // set (carrier) Frequency MID
  rfmWriteReg(0x09, 0x80); // set (carrier) Frequency LSB
  rfmWriteReg(0x03, 0x26); // set Bitrate MSB
  rfmWriteReg(0x04, 0xF0); // set Bitrate LSB
}


void sendFunc1(int debugPulses, byte byte1, byte byte2, byte byte3)
{
  // PRECONDITION: We're already in TX mode
  // send first block (looks fine (one (low-)bit delay at start))
  debugPulse(debugPulses);
  rfmWriteReg(0x01, 0x0C); //  set OP-MODE (TX)
  rfmWriteReg(0x00, byte1); //  FILL FIFO
  rfmWriteReg(0x00, byte2); //  FILL FIFO
  rfmWriteReg(0x00, byte3); //  FILL FIFO
  
  // wait until it is definitely completed (we can't use PacketSent because we don't leave TX mode here)
  delay(13);
}


void sendFunc2(int debugPulses, byte byte1, byte byte2, byte byte3)
{
  // send block of 3 bytes, pre-fill FIFO and change to TX Mode
  rfmWriteReg(0x00, byte1); // FILL FIFO
  rfmWriteReg(0x00, byte2); // FILL FIFO
  rfmWriteReg(0x00, byte3); // FILL FIFO
  debugPulse(debugPulses);
  rfmWriteReg(0x01, 0x0C); // set OP-MODE (TX)

  // wait for the transmission to finish and go to standby
  while ((rfmReadReg(0x28) & 0x08) == 0) 
  {
    delayMicroseconds(200);
  }
  setOperationMode(MODE_STANDBY);
  delay(5);
}

void sendFunc3(int debugPulses, byte byte1, byte byte2, byte byte3)
{
  // send block of 3 bytes, pre-fill FIFO and change to TX Mode
  rfmWriteReg(0x00, byte1); // FILL FIFO
  rfmWriteReg(0x00, byte2); // FILL FIFO
  rfmWriteReg(0x00, byte3); // FILL FIFO
  debugPulse(debugPulses);
  rfmWriteReg(0x01, 0x0C); //  set OP-MODE (TX)

  // wait for the transmission to finish and go to standby (with going to SLEEP briefly before)
  while ((rfmReadReg(0x28) & 0x08) == 0) 
  {
    delayMicroseconds(200);
  }
  setOperationMode(MODE_SLEEP);
  setOperationMode(MODE_STANDBY);
  delay(4);
}


void testCase1()
{
  setOperationMode(MODE_TX);
  sendFunc1(3, 0x3C, 0xAA, 0x0C);
  sendFunc1(1, 0x3C, 0xAA, 0x0D);
  sendFunc1(1, 0x3C, 0xAA, 0x0C);

  setOperationMode(MODE_SLEEP);
  setOperationMode(MODE_STANDBY);
}

void testCase2()
{
  sendFunc2(4, 0x3C, 0xAA, 0x0C);
  sendFunc2(1, 0x3C, 0xAA, 0x0D);
  sendFunc2(1, 0x3C, 0xAA, 0x0C);

  setOperationMode(MODE_SLEEP);
  setOperationMode(MODE_STANDBY);
}

void testCase3()
{
  sendFunc3(5, 0x3C, 0xAA, 0x0C);
  sendFunc3(1, 0x3C, 0xAA, 0x0D);
  sendFunc3(1, 0x3C, 0xAA, 0x0C);

  setOperationMode(MODE_SLEEP);
  setOperationMode(MODE_STANDBY);
}




void setup()
{
  Serial.begin(SERIAL_BAUD);

  pinMode(PIN_CS_RFM, OUTPUT);      // ATTENTION: PIN-10 (aka. PB2/SS) needs to be set to output before initializing SPI
  digitalWrite(PIN_CS_RFM, HIGH);   // HIGH means "NOT SELECTED"
  pinMode(PIN_CS_FLASH, OUTPUT);
  digitalWrite(PIN_CS_FLASH, HIGH);
  pinMode(PIN_DEBUG, OUTPUT);
  digitalWrite(PIN_DEBUG, LOW);

  // some time for RFM to settle
  delay(100);

  SPI.begin();
  SPI.beginTransaction(SPISettings(SPI_BAUD, MSBFIRST, SPI_MODE0));  // 8mhz would be maximum (set way lower, so we can run sikrok slower)

  rfmWriteReg(0x01, 0x04); // set OP-MODE (STANDBY)

}


void loop()
{
  int action;

  Serial.println("ready for next action:");
  while (Serial.available() == 0)
    ;
    
  action = Serial.parseInt();
  
  if (action == 1)
  {
   debugPulse(2);
   initialize_OOK();
   delay(15);
   testCase1();   
   delay(40);
   testCase2();   
   delay(40);
   testCase3();   
  }  
  else if (action == 2)
  {
    initialize_OOK();
  }
  else if (action == 4)
  {
    testCase1();
  }
  else if (action == 5)
  {
    testCase2();
  }
  else if (action == 6)
  {
    testCase3();
  }
  else 
  {
    Serial.print("unknown action: ");
    Serial.println(action);
  }
}
